#include "sink.h"
#include "stream.h"
#include "aggr.h"

#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

static sink_t *sinks;
int num_sinks;

int sink_hash(uint8_t *key, unsigned int len) {
   unsigned int b = 378551;
   unsigned int a = 63689;
   unsigned int hash = 0;
   unsigned int i = 0;

   for(i = 0; i < len; key++, i++) {
      hash = hash * a + (*key);
      a = a * b;
   }

   return hash;
}

void sink_init(aggregator_t *aggregator, int num_nodes) {
  num_sinks = num_nodes;
  sinks = (sink_t *)malloc(sizeof(sink_t) * num_sinks);

  for(int i = 0; i < num_sinks; i++) {
    const node_t *n = &nodes[i];
    sinks[i].fd = sink_connect(n->address, n->port);
    sinks[i].buffer = (char *)malloc(BUFFER_SIZE);
    sinks[i].buffer_pos = 0;
    sinks[i].slab = aggregator->slab;
    sinks[i].slab_count = aggregator->count % AGGREGATE_SLAB_SIZE;
    sinks[i].slab_pos = 0;
  }
}

int sink_connect(char *address, int port) {
  struct sockaddr_in addr;
  int fd;

  if((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    fprintf(stderr, "Couldn't create sink socket\n");
    exit(1);
  }

  memset(&addr, 0, sizeof(addr));
  addr.sin_port = htons(port);
  addr.sin_family = AF_INET;
  if(!inet_pton(AF_INET, address, (struct in_addr *)&addr.sin_addr.s_addr)) {
    fprintf(stderr, "Invalid IP address: %s\n", address);
    exit(1);
  }

  if(connect(fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
    fprintf(stderr, "Couldn't connect to %s:%d\n", address, port);
    exit(1);
  }

  return fd;
}

void sink_add(int fd, aggregate_slab_t *slab, size_t slab_count) {
  for(int i = 0; i < num_sinks; i++) {
    if(sinks[i].fd == -1) {
      sinks[i].fd = fd;
      sinks[i].buffer = (char *)malloc(BUFFER_SIZE);
      sinks[i].buffer_pos = 0;

      sinks[i].slab = slab;
      sinks[i].slab_count = slab_count;
      sinks[i].slab_pos = 0;
      break;
    }
  }
}

void sink_close(sink_t *sink) {
  fprintf(stderr, "[%i] Closing connection\n", sink->fd);
  close(sink->fd);
  sink->fd = -1;
  sink->buffer_pos = 0;
}

int sink_add_fds(fd_set *fds) {
  int max = -1;
  for(int i = 0; i < num_sinks; i++) {
    if(sinks[i].fd != -1) {
      FD_SET(sinks[i].fd, fds);
      if(sinks[i].fd > max) {
        max = sinks[i].fd;
      }
    }
  }
  return max;
}

void sink_process(fd_set *fds, aggregator_t *aggregator) {
  for(int i = 0; i < num_sinks; i++) {
    sink_t *s = &sinks[i];

    if(s->fd == -1 || !FD_ISSET(s->fd, fds))
      continue;


    /* Send off whatever is buffered */
    ssize_t len = write(s->fd, s->buffer, s->buffer_pos);

    /* Shift leftover data to beginning of buffer */
    memmove(s->buffer, s->buffer + len, s->buffer_pos - len);
    s->buffer_pos -= len;

    /* Process slabs, fill unused buffer space with aggregates */

    /* Calculate how many aggregates we can fit in the buffer */
    size_t room = (BUFFER_SIZE - s->buffer_pos) / (sizeof(aggregate_t) + aggregator->key_length);

    /* While there is buffer space and aggregates available */
    while(room > 0 && s->slab != NULL) {
      void *a = s->slab->aggregates + (sizeof(aggregate_t) + aggregator->key_length) * s->slab_pos;

      /* Check if aggregate belongs to this node */
      uint8_t *key = KEY(a);
      unsigned int hash = sink_hash(key, aggregator->key_length);

      if(hash % num_sinks == i) {
        /* Write aggregate to buffer */
        memcpy(s->buffer + s->buffer_pos, a, sizeof(aggregate_t) + aggregator->key_length);
        s->buffer_pos += sizeof(aggregate_t) + aggregator->key_length;
        room--;
      }

      s->slab_pos++;
      if(s->slab_pos >= s->slab_count) {
        s->slab = s->slab->next;
        s->slab_count = AGGREGATE_SLAB_SIZE;
      }
    }

    /* Close connection if no more data was added and we have sent it all */
    if(s->buffer_pos == 0 && s->slab == NULL) {
      sink_close(s);
      continue;
    }

  }
}
